# Generated by Django 5.0.7 on 2024-07-26 05:25

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="FileManager",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=255, verbose_name="Name")),
                ("file", models.FileField(upload_to="")),
                (
                    "status",
                    models.CharField(
                        choices=[
                            ("COMPLETED", "Completed"),
                            ("ERROR", "Error"),
                            ("PENDING", "Pending"),
                            ("PROCESSING", "Processing"),
                        ],
                        default="PENDING",
                        max_length=255,
                        verbose_name="Status",
                    ),
                ),
                ("checksum", models.CharField(max_length=255, unique=True)),
                ("eof", models.BooleanField(default=False)),
                ("metadata", models.JSONField(default=dict)),
            ],
            options={
                "verbose_name": "Django Chunk File Upload",
                "verbose_name_plural": "Django Chunk File Upload",
                "db_table": "django_chunk_file_upload",
                "indexes": [
                    models.Index(fields=["checksum"], name="file_manager_checksum_idx")
                ],
            },
        ),
    ]



from __future__ import annotations

from django.contrib import admin

from .forms import ChunkedUploadFileAdminForm
from .models import FileManager


@admin.register(FileManager)
class FileManagerModelAdmin(admin.ModelAdmin):
    form = ChunkedUploadFileAdminForm
    list_display = (
        "id",
        "name",
        "status",
        "created_at",
        "updated_at",
    )

    add_form_template = "django_chunk_file_upload/admin/add_form.html"
    change_form_template = "django_chunk_file_upload/admin/change_form.html"



from __future__ import annotations

from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from .constants import StatusChoices, TypeChoices


class FileManagerMixin(models.Model):
    """File Manager Mixin for Django Models"""

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)
    file = models.FileField()
    status = models.CharField(
        _("Status"),
        max_length=255,
        choices=StatusChoices.choices,
        default=StatusChoices.PENDING,
    )
    type = models.CharField(
        _("Type"),
        max_length=255,
        choices=TypeChoices.choices,
        default=TypeChoices.__empty__,
    )
    checksum = models.CharField(max_length=255)
    eof = models.BooleanField(default=False)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="files",
    )
    metadata = models.JSONField(default=dict)

    class Meta:
        abstract = True

    def __str__(self):
        return self.name

    @property
    def name(self) -> str:
        if "name" in self.metadata and self.metadata["name"]:
            return self.metadata["name"]
        return self.file.name


class FileManager(FileManagerMixin):
    """File Manager for Django Models"""

    class Meta:
        db_table = "django_chunk_file_upload"
        indexes = [models.Index(fields=["checksum"], name="%(class)s_checksum_idx")]
        ordering = ("-created_at",)
        unique_together = (
            "user",
            "checksum",
        )
        verbose_name = _("File Manager")
        verbose_name_plural = _("File Manager")



from __future__ import annotations


class BasePermission:
    safe_methods = ("GET", "POST", "DELETE")

    def has_permission(self, request, view) -> bool:
        return False

    def is_safe_method(self, request, view) -> bool:
        return bool(
            str(request.method).upper() in [m.upper() for m in self.safe_methods]
        )


class AllowAny(BasePermission):
    def has_permission(self, request, view) -> bool:
        if self.is_safe_method(request, view):
            return True
        return False


class IsAuthenticated(BasePermission):
    def has_permission(self, request, view) -> bool:
        if self.is_safe_method(request, view):
            return bool(request.user and request.user.is_authenticated)
        return False


class IsSuperUser(BasePermission):
    def has_permission(self, request, view) -> bool:
        if self.is_safe_method(request, view):
            return bool(request.user and request.user.is_superuser)
        return False


class IsAdminUser(BasePermission):
    def has_permission(self, request, view) -> bool:
        if self.is_safe_method(request, view):
            return bool(
                request.user and (request.user.is_staff or request.user.is_superuser)
            )
        return False


class IsAuthenticatedOrReadOnly(BasePermission):
    def has_permission(self, request, view) -> bool:
        if self.is_safe_method(request, view):
            return bool(
                request.user
                and (request.user.is_authenticated or request.user.is_superuser)
            )
        return False



from __future__ import annotations

from django import forms

from .models import FileManager
from .widgets import DragDropFileField, MultipleDragDropFileField


class ChunkedUploadFileForm(forms.ModelForm):
    file = MultipleDragDropFileField(label="", required=True)

    class Meta:
        model = FileManager
        fields = ("file",)
        exclude = (
            "checksum",
            "eof",
            "status",
            "metadata",
        )


class ChunkedUploadFileAdminForm(ChunkedUploadFileForm):
    file = DragDropFileField()



from __future__ import annotations

from django.urls import path

from .views import ChunkedUploadView


app_name = "django_chunk_file_upload"
urlpatterns = [
    path("uploads/", ChunkedUploadView.as_view(), name="uploads"),
]



from __future__ import annotations

from django.db import IntegrityError
from django.db.models import ManyToManyField, QuerySet
from django.http import Http404, JsonResponse
from django.utils.translation import gettext_lazy as _
from django.views.generic.edit import FormView

from .app_settings import app_settings
from .constants import ActionChoices
from .forms import ChunkedUploadFileForm
from .models import FileManager
from .typed import (
    ArchiveFile,
    AudioFile,
    BinaryFile,
    DocumentFile,
    File,
    FontFile,
    HyperTextFile,
    ImageFile,
    JSONFile,
    MicrosoftExcelFile,
    MicrosoftPowerPointFile,
    MicrosoftWordFile,
    SeparatedFile,
    XMLFile,
)
from .utils import get_md5_checksum


class ChunkedUploadView(FormView):
    """Chunked upload view."""

    http_method_names = ["get", "post", "delete"]
    chunk_size = app_settings.chunk_size
    file_class = File
    file_status = app_settings.status
    form_class = ChunkedUploadFileForm
    optimize = app_settings.optimize
    permission_classes = app_settings.permission_classes
    remove_file_on_update = app_settings.remove_file_on_update
    template_name = "django_chunk_file_upload/chunked_upload.html"
    upload_to = app_settings.upload_to

    def check_object_permissions(self, request):
        for permission in self.permission_classes:
            permission = permission() if isinstance(permission, type) else permission
            if permission.has_permission(request, self):
                return True
        return False

    def has_add_permission(self, request, obj=None) -> bool:
        return self.check_object_permissions(request)

    def has_view_permission(self, request, obj=None) -> bool:
        return self.check_object_permissions(request)

    def has_change_permission(self, request, obj=None) -> bool:
        return self.check_object_permissions(request)

    def has_delete_permission(self, request, obj=None) -> bool:
        return self.check_object_permissions(request)

    def is_valid(self, form, file_obj) -> bool:
        if form.is_valid() and file_obj.is_valid():
            return True
        return False

    def get_model(self):
        return self.form_class.Meta.model

    def get_instance(self):
        opts = dict(
            user=self.request.user if self.request.user.is_authenticated else None
        )
        pk = self.request.headers.get("x-file-id")
        if pk:
            opts["pk"] = pk
            return self.get_model().objects.filter(**opts).first()

        if self.request.headers.get("x-file-checksum"):
            opts["checksum"] = self.request.headers["x-file-checksum"]
            return self.get_model().objects.filter(**opts).first()

    def get_context_data(self, **kwargs):
        context = super(ChunkedUploadView, self).get_context_data(**kwargs)
        context["chunk_size"] = self.chunk_size
        return context

    def get(self, request, *args, **kwargs):
        return self._get(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """Override POST method from View."""

        action = request.POST.get("action")
        if action == ActionChoices.UPDATE:
            return self._update(request, *args, **kwargs)
        return self._post(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        """Override DELETE method from View."""

        return self._delete(request, *args, **kwargs)

    def _get(self, request, *args, **kwargs):
        if self.has_view_permission(request):
            return super(ChunkedUploadView, self).get(request, *args, **kwargs)
        raise Http404

    def _post(self, request, *args, **kwargs):
        form, file_obj = self._get_form_file(request, *args, **kwargs)
        if self.has_add_permission(self.request) and self.is_valid(form, file_obj):
            instance = self.get_instance()
            return self.chunked_upload(instance, form, file_obj)

        file_obj.message = _("Permission denied.")
        return self.ajax_response(None, file_obj, status=400, save=False)

    def _update(self, request, *args, **kwargs):
        form, file_obj = self._get_form_file(request, *args, **kwargs)
        if self.has_change_permission(self.request) and self.is_valid(form, file_obj):
            instance = self.get_instance()
            if instance:
                if self.remove_file_on_update:
                    instance.file.delete()
                instance.eof = False
                instance.save()
                return self.chunked_upload(instance, form, file_obj)

            file_obj.message = _("Not found.")
            return self.ajax_response(None, file_obj, status=400, save=False)

        file_obj.message = _("Permission denied.")
        return self.ajax_response(None, file_obj, status=400, save=False)

    def _delete(self, request, *args, **kwargs):
        form, file_obj = self._get_form_file(request, *args, **kwargs)
        if self.has_delete_permission(request):
            instance = self.get_instance()
            if instance and (
                self.request.user.is_superuser or self.request.user == instance.user
            ):
                instance.file.delete()
                instance.delete()
                file_obj.message = _("The file deleted successfully.")
                return self.ajax_response(None, file_obj, status=200, save=False)

        file_obj.message = _("Permission denied.")
        return self.ajax_response(None, file_obj, status=400, save=False)

    def _get_form_file(
        self, request, *args, **kwargs
    ) -> tuple[ChunkedUploadFileForm, File]:
        form = self.get_form(self.form_class)
        file_obj = self.file_class.from_request(self.request, self.upload_to)
        return form, file_obj

    def chunked_upload(self, instance, form, file_obj):
        """Chunked upload file

        Handle requests from JQuery AJAX, save files to server.
        Check MD5 checksum of file when upload is complete, if not correct delete file from server.

        Args:
            instance (FileManager): FileManager object.
            form (ChunkedUploadFileForm): Chunked upload from.
            file_obj (File): File metadata instance.

        Returns:
            JsonResponse: return file metadata data.
        """
        if not instance:
            instance = form.instance

        if instance.eof:
            file_obj.message = _("The file already exists.")
            return self.ajax_response(instance, file_obj, 403, save=False)

        kwargs, m2m_kwargs = self.get_kwargs(form)
        for k, v in kwargs.items():
            setattr(instance, k, v)

        try:
            self.save(instance, file_obj)
            self.save_m2m(instance, **m2m_kwargs)
            file_obj.write("ab+" if instance.file else "wb+")
        except IntegrityError as e:
            return self.raise_exception(e, instance, file_obj)

        except Exception as e:
            return self.raise_exception(e, instance, file_obj)

        if file_obj.eof is False:
            return self.ajax_response(instance, file_obj)

        checksum = get_md5_checksum(file_obj.save_path)
        if checksum != file_obj.checksum:
            instance.file.delete()
            instance.eof = False
            file_obj.message = _("The file does not match the MD5 checksum.")
            return self.ajax_response(instance, file_obj, 400)

        self.background_task(instance)
        if self.optimize:
            file_obj.optimize(instance)
        return self.ajax_response(instance, file_obj)

    def raise_exception(
        self, exception: Exception, instance: FileManager, file_obj: File
    ):
        error = str(exception)
        file_obj.message = error
        if isinstance(exception, IntegrityError):
            file_obj.message = _("DB error: %s.") % error
            if "UNIQUE" in error:
                file_obj.message = _("The file was created by another user.")

            return self.ajax_response(instance, file_obj, 400, False)

        return self.ajax_response(instance, file_obj, 400)

    def ajax_response(
        self,
        instance: FileManager,
        file_obj: File,
        status: int = 201,
        save: bool = True,
    ):
        if save:
            self.save(instance, file_obj)

        data = file_obj.to_response()
        if instance and instance.eof:
            data["url"] = instance.file.url

        return JsonResponse(
            data=data,
            status=status,
        )

    def get_kwargs(self, form, **kwargs) -> tuple[dict, dict]:
        kwargs, m2m_kwargs = {}, {}
        m2m_field = {
            field.attname: []
            for field in form.instance._meta.get_fields()
            if isinstance(field, ManyToManyField)
        }
        if hasattr(form, "cleaned_data"):
            for k, v in form.cleaned_data.items():
                if k in m2m_field:
                    m2m_kwargs[k] = v
                else:
                    kwargs[k] = v
        return kwargs, m2m_kwargs

    def save_m2m(self, instance, **kwargs):
        for field, values in kwargs.items():
            m2m_field = getattr(instance, field, None)
            if m2m_field:
                if isinstance(values, QuerySet):
                    removed = []
                    added = {obj.id: obj for obj in values}
                    for obj in m2m_field.all():
                        if obj.id not in added:
                            removed.append(obj)
                        else:
                            added.pop(obj.id, None)

                    m2m_field.add(*added.values())
                    m2m_field.remove(*removed)
                else:
                    m2m_field.clear()

    def save(self, instance: FileManager, file_obj: File):
        instance.eof = file_obj.eof
        instance.file = file_obj.path
        instance.type = file_obj.type
        instance.user = file_obj.user
        instance.status = self.file_status
        if not instance.checksum:
            instance.checksum = file_obj.checksum

        if app_settings.is_metadata_storage:
            instance.metadata = file_obj.to_metadata()

        instance.save()

    def background_task(self, instance):
        pass


class ChunkArchiveUploadView(ChunkedUploadView):
    """Chunk Archive Upload View"""

    file_class = ArchiveFile


class ChunkAudioUploadView(ChunkedUploadView):
    """Chunk Audio Upload View"""

    file_class = AudioFile


class ChunkBinaryUploadView(ChunkedUploadView):
    """Chunk Binary Upload View"""

    file_class = BinaryFile


class ChunkDocumentUploadView(ChunkedUploadView):
    """Chunk Document Upload View"""

    file_class = DocumentFile


class ChunkFontUploadView(ChunkedUploadView):
    """Chunk Font Upload View"""

    file_class = FontFile


class ChunkHyperTextUploadView(ChunkedUploadView):
    """Chunk HyperText Upload View"""

    file_class = HyperTextFile


class ChunkImageUploadView(ChunkedUploadView):
    """Chunk Image Upload View"""

    file_class = ImageFile


class ChunkJSONUploadView(ChunkedUploadView):
    """Chunk JSON Upload View"""

    file_class = JSONFile


class ChunkMicrosoftWordUploadView(ChunkedUploadView):
    """Chunk Microsoft Word Upload View"""

    file_class = MicrosoftWordFile


class ChunkMicrosoftPowerPointUploadView(ChunkedUploadView):
    """Chunk Microsoft PowerPoint Upload View"""

    file_class = MicrosoftPowerPointFile


class ChunkMicrosoftExcelUploadView(ChunkedUploadView):
    """Chunk Microsoft Excel Upload View"""

    file_class = MicrosoftExcelFile


class ChunkSeparatedUploadView(ChunkedUploadView):
    """Chunk Separated Upload View"""

    file_class = SeparatedFile


class ChunkXMLUploadView(ChunkedUploadView):
    """Chunk XML Upload View"""

    file_class = XMLFile



from __future__ import annotations

import importlib
from dataclasses import dataclass, field, fields

from django.conf import settings

from . import permissions
from .constants import StatusChoices


@dataclass(kw_only=True)
class _Settings:

    @classmethod
    def get_kwargs(cls, **kwargs) -> dict:
        model_fields = {field.name for field in fields(cls)}
        return {k: v for k, v in kwargs.items() if k in model_fields}

    @classmethod
    def from_kwargs(cls, **kwargs) -> "_Settings":
        return cls(**cls.get_kwargs(**kwargs))


@dataclass(kw_only=True)
class _ImageSettings(_Settings):
    quality: int = 82
    compress_level: int = 9
    max_width: int = 1280
    max_height: int = 720
    to_webp: bool = True


@dataclass(kw_only=True)
class _LazySettings(_Settings):
    css: tuple | list | set = (
        "css/upload.chunk.css",
        "css/toastr.min.css",
        "css/sweetalert2.min.css",
    )
    js: tuple | list | set = (
        "js/jquery-3.7.1.min.js",
        "js/spark-md5.min.js",
        "js/toastr.min.js",
        "js/sweetalert2.min.js",
        "js/upload.chunk.js",
    )
    upload_to: str = "%Y/%m/%d"
    chunk_size: int = 1024 * 1024 * 2  # 2MB
    is_metadata_storage: bool = False
    remove_file_on_update: bool = True
    status: StatusChoices = StatusChoices.PENDING
    permission_classes: tuple[permissions.BasePermission] = (
        permissions.IsAuthenticated,
    )
    optimize: bool = True
    image_optimizer: _ImageSettings = field(default_factory=_ImageSettings)

    @classmethod
    def from_kwargs(cls, **kwargs) -> "_LazySettings":
        kwargs = cls.get_kwargs(**kwargs)
        js = kwargs.pop("js", None) or []
        if js and isinstance(js, list):
            kwargs["js"] = list(set(js + ["js/upload.chunk.js"]))

        image_optimizer = kwargs.pop("image_optimizer", {}) or {}
        if image_optimizer and isinstance(image_optimizer, dict):
            kwargs["image_optimizer"] = _ImageSettings.from_kwargs(**image_optimizer)

        permission_classes = kwargs.pop("permission_classes")
        if permission_classes and isinstance(
            permission_classes, (tuple, list, set, str)
        ):
            if isinstance(permission_classes, str):
                permission_classes = [permission_classes]

            perms = []
            for permission_class in permission_classes:
                paths = permission_class.split(".")
                module = importlib.import_module(".".join(paths[:-1]), "")
                permission_class = getattr(module, paths[-1])
                perms.append(permission_class)

            kwargs["permission_classes"] = tuple(perms)
        return cls(**kwargs)


app_settings = _LazySettings.from_kwargs(
    **getattr(settings, "DJANGO_CHUNK_FILE_UPLOAD", {})
)



from __future__ import annotations

from django.db.models import TextChoices
from django.utils.translation import gettext_lazy as _


class TypeChoices(TextChoices):
    ARCHIVE = "ARCHIVE", _("ARCHIVE")
    AUDIO = "AUDIO", _("AUDIO")
    BINARY = "BINARY", _("BINARY")
    DOCUMENT = "DOCUMENT", _("DOCUMENT")
    FONT = "FONT", _("FONT")
    HYPERTEXT = "HYPERTEXT", _("HYPERTEXT")
    IMAGE = "IMAGE", _("IMAGE")
    JSON = "JSON", _("JSON")
    MICROSOFT_EXCEL = "MICROSOFT_EXCEL", _("MICROSOFT_EXCEL")
    MICROSOFT_POWERPOINT = "MICROSOFT_POWERPOINT", _("MICROSOFT_POWERPOINT")
    MICROSOFT_WORD = "MICROSOFT_WORD", _("MICROSOFT_WORD")
    SOURCE_CODE = "SOURCE_CODE", _("SOURCE_CODE")
    SEPARATED = "SEPARATED", _("SEPARATED")
    TEXT = "TEXT", _("TEXT")
    VIDEO = "VIDEO", _("VIDEO")
    XML = "XML", _("XML")
    __empty__ = _("Unknown")


class StatusChoices(TextChoices):
    COMPLETED = "COMPLETED", _("Completed")
    ERROR = "ERROR", _("Error")
    PENDING = "PENDING", _("Pending")
    PROCESSING = "PROCESSING", _("Processing")


class ActionChoices(TextChoices):
    CREATE = "_add", _("Add")
    UPDATE = "_save", _("Save")
    DELETE = "_delete", _("Delete")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
django-chunk-file-upload
------------

Tests for `django-chunk-file-upload` models module.
"""

from django.test import TestCase


class TestDjangoChunkFileUpload(TestCase):

    def setUp(self):
        pass

    def test_something(self):
        pass

    def tearDown(self):
        pass
